/*
 * _mssql module
 *
 * Joon-cheol Park
 * jooncheol@gmail.com
 * http://www.exman.pe.kr
 *
 * thankyou: 
 * 		Hans Roh (hans@lufex.com) - List ref count patch 2004.04.09
 *
 * */
#include "Python.h"

#ifdef MS_WIN32
	#define	DBNTWIN32		// must identify operating system environment
	#include "windows.h"
	#include "odbcinst.h"
	#include <sqlfront.h>
	#include <sqldb.h>	   // DB-LIB header file (should always be included)
#else
	#include <sybfront.h>
	#include <sybdb.h>	   // DB-LIB header file (should always be included)
	#include <dblib.h>	   // DB-LIB header file (should always be included)

	#define SQLNUMERIC    SYBNUMERIC    
	#define SQLDECIMAL    SYBDECIMAL
	#define SQLBIT        SYBBIT
	#define SQLINT1       SYBINT1
	#define SQLINT2       SYBINT2
	#define SQLINT4       SYBINT4
	#define SQLFLT4       SYBREAL
	#define SQLFLT8       SYBFLT8
	#define SQLDATETIME   SYBDATETIME
	#define SQLMONEY      SYBMONEY
	#define SQLVARBINARY  SYBVARBINARY
	#define SQLBINARY     SYBBINARY
	#define SQLIMAGE      SYBIMAGE
	#define SQLDATETIME      SYBDATETIME
	#define SQLDATETIME4     SYBDATETIME4

	#define BYTE 		  unsigned char
	#define LPBYTE 		  unsigned char*
#endif

#include <stdio.h>

#include <string.h>     // include for string functions
#include <stdlib.h>     // include for malloc and free


static PyObject *_mssql_error;

// Prototypes for internal functions.
int DetermineRowSize(DBPROCESS *,int);
PyObject *GetRow ( DBPROCESS *dbproc);

int err_handler();
int msg_handler();

PyObject *GetHeaders ( DBPROCESS *dbproc);

// dbproc
typedef struct
{
	PyObject_HEAD
	//PDBPROCESS dbproc;
	DBPROCESS *dbproc;
	int connected;
} PY_DBPROC;
extern PyTypeObject PY_DBPROC_Type;

#define MSSQLDB_MSGSIZE 1024
static char _mssql_query_error[MSSQLDB_MSGSIZE]={0,};
static char _mssql_query_message[MSSQLDB_MSGSIZE]={0,};



//int err_handler(PDBPROCESS dbproc, int severity, int dberr, int oserr, char * dberrstr, char * oserrstr)
int err_handler(dbproc, severity, dberr, oserr, dberrstr, oserrstr)
	DBPROCESS       *dbproc;
	int             severity;
	int             dberr;
	int             oserr;
	char            *dberrstr;
	char            *oserrstr;
{
	if (dberrstr != NULL)
	{
		sprintf(_mssql_query_error,"DB-LIB Error - %s\n",dberrstr);
	}

	if (oserr != DBNOERR)
	{
		sprintf(_mssql_query_error,"Operating-system Error - %s\n",oserrstr);
	}


	if ((dbproc == NULL) ||	(DBDEAD(dbproc)))
		return(INT_EXIT);
	else
		return(INT_CANCEL);
}

//int msg_handler(PDBPROCESS dbproc, DBINT msgno, int msgstate, int severity, char * msgtext)
int msg_handler(dbproc, msgno, msgstate, severity, msgtext,
		                srvname, procname, line)

DBPROCESS       *dbproc;
DBINT           msgno;
int             msgstate;
int             severity;
char            *msgtext;
char            *srvname;
char            *procname;
DBUSMALLINT     line;

{
	sprintf(_mssql_query_message,"SQL Server message %ld, state %d, severity %d:\n\t%s\n", msgno, msgstate, severity, msgtext);
	return(0);
}



static PyObject *_mssql_connect(self, args)
	PY_DBPROC *self;
    PyObject *args;
{ 
	PY_DBPROC *dbprochandle;
	//PLOGINREC  login;   // allocate	a DB-LIB login structure
	LOGINREC  *login;   // allocate	a DB-LIB login structure
	char *ServerName, *User, *Passwd;
	long port=1433;

#ifdef MS_WIN32
	char	szAttribute[1024];
	char	*p = szAttribute;
#endif


	if(!PyArg_ParseTuple(args,"sss|l",&ServerName,&User,&Passwd,&port))
	{
		PyErr_SetString(_mssql_error,"Usage: connect(ServerName,User,Passwd,port=1433)");
		return NULL;
	}


#ifdef MS_WIN32
	wsprintf(szAttribute, "DSN=pymssql;SERVER=%s;Trusted_connection=No;Network=DBMSSOCN;;", ServerName);
	while(*p)
	{
		if(*p == ';') *p = '\0';
		p++;
	}
			
	if(SQLConfigDataSource(NULL, ODBC_ADD_DSN, "SQL Server", szAttribute) == FALSE)
	{
		PyErr_SetString(_mssql_error,"DSN ADD ERROR");
		return NULL;
	}
#endif

	


#ifdef MS_WIN32
	if(dbinit() == (char *)NULL)
	{
		PyErr_SetString(_mssql_error,"Commucations layer not loaded");
		return NULL;
	}
#else
	if(dbinit() == FAIL)
	{
		PyErr_SetString(_mssql_error,"Commucations layer not loaded");
		return NULL;
	}
#endif

    dbsettime(30);          // set timeouts to 30 seconds

	login = dblogin();  // get login record from DB-LIB

	DBSETLUSER(login, User);
	DBSETLPWD(login, Passwd);
	DBSETLAPP(login, "pymssql");
#ifdef MS_WIN32
	DBSETLVERSION(login, DBVER60);
#else
	DBSETLHOST(login, ServerName);
#endif

	dbprochandle = PyObject_NEW(PY_DBPROC, &PY_DBPROC_Type);
	if(dbprochandle == NULL)
	{
		PyErr_SetString(_mssql_error,"Not Created  PY_DBPROC Object");
		return NULL;
	}

#ifdef MS_WIN32
	dberrhandle((DBERRHANDLE_PROC)err_handler);
	dbmsghandle((DBMSGHANDLE_PROC)msg_handler);
#else
	dberrhandle(err_handler);
	dbmsghandle(msg_handler);
	// Set port
	tds_set_port(login->tds_login, port);
#endif


	if((dbprochandle->dbproc=dbopen(login,ServerName)) == NULL)
	{
		PyErr_SetString(_mssql_error,"Can't Connecting MSSQL Server");
		return NULL;
	}


	//dbsetopt(dbprochandle->dbproc, DBBUFFER, "2", -1);
	//dbsetopt(dbprochandle->dbproc, DBTEXTLIMIT, "8096", -1);

#ifdef MS_WIN32
	dbfreelogin(login); // Frees a login record.
#else
	dbloginfree(login); // Frees a login record.
#endif

	dbprochandle->connected=1;
	return (PyObject *)dbprochandle;
}






static PyMethodDef _mssql_methods[] = {
    {"connect",  _mssql_connect, METH_VARARGS},
    {NULL,      NULL}        /* Sentinel */
};


/* platform   ֱ  */
#ifdef MS_WIN32
__declspec(dllexport)
#endif

void init_mssql()
{
    PyObject *m, *d;

    m = Py_InitModule("_mssql", _mssql_methods);
    d = PyModule_GetDict(m);
    _mssql_error = PyErr_NewException("_mssql.error", NULL, NULL);
    PyDict_SetItemString(d, "error", _mssql_error);
} 



// fetch_array
static PyObject *_mssql_fetch_array(PY_DBPROC *self, PyObject *args)
{
	PyObject *resultSet=NULL;
	PyObject *o=NULL;
	int retc;
	int rows;
	int i;
	//  Ʈ ʱȭ
	//  
	// resultSet[ result(  header(...), recored(...) ),
	// 			  result(...),
	// 			  ...]
	if (!(resultSet = PyList_New(0)))
	{
		PyErr_SetString(_mssql_error,"Not Created Fetch Tuple");
		return NULL;
	}
	if(!PyArg_NoArgs(args))
	{
		PyErr_SetString(_mssql_error, "This Method is not required Arguments");
		return NULL;
	}

	if(!self->connected)
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}

	// command executed correctly, get results information
	while((retc = dbresults(self->dbproc)) != NO_MORE_RESULTS)
	{
		PyObject *result=NULL;
		PyObject *header=NULL;
		PyObject *record=NULL;

		if (retc == FAIL)           // if error get out of loop
			break;

		//   ʱȭ
		if (!(result = PyTuple_New(3)))
		{
			PyErr_SetString(_mssql_error,"Not Created Result Tuple");
			return NULL;
		}
		// Į  ʱȭ
		header = GetHeaders(self->dbproc);       // print header data
		PyTuple_SET_ITEM(result,0,header);

		// Ʈ  ʱȭ
		if (!(record = PyList_New(0)))
		{
			PyErr_SetString(_mssql_error,"Not Created Record Tuple");
			return NULL;
		}
		// loop on each row, until all read
		while((retc= dbnextrow(self->dbproc))!=NO_MORE_ROWS)
		{
			if(retc == FAIL)        // if fail, then clear
			{                   // connection completely, just
				dbcancel(self->dbproc);   // in case.
				//PyTuple_SET_ITEM(record,i++,Py_BuildValue("s",Py_BuildValue("s","\0")));
				o = Py_BuildValue("s", "");
				PyList_Append(record, o);
				Py_DECREF(o);
				break;
			}
			else
			{
				o = GetRow(self->dbproc);
				PyList_Append(record,o);
				Py_DECREF(o);
			}
		}
		rows=DBCOUNT(self->dbproc);
		i=0;
		// affected_rows 
		PyTuple_SET_ITEM(result,1,Py_BuildValue("i",rows));
		PyTuple_SET_ITEM(result,2,record);

		PyList_Append(resultSet,result);
		Py_DECREF(result);
	}   // end while(dbresults())

	return resultSet;
}





// query
static PyObject *_mssql_query(PY_DBPROC *self, PyObject *args)
{
	char *Query;
	if(!PyArg_ParseTuple(args,"s",&Query))
	{
		PyErr_SetString(_mssql_error,"Usage: query(Query)");
		return NULL;
	}

	if(!self->connected)
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}



#ifdef MS_WIN32
	ZeroMemory(_mssql_query_error,MSSQLDB_MSGSIZE);
	ZeroMemory(_mssql_query_message,MSSQLDB_MSGSIZE);
#else
	memset(_mssql_query_error,0,MSSQLDB_MSGSIZE);
	memset(_mssql_query_message,0,MSSQLDB_MSGSIZE);
#endif

	dbcmd(self->dbproc,Query);
	if(dbsqlexec(self->dbproc))
	{
		//dbfreebuf(self->dbproc);
		return Py_BuildValue("i",1);
	}
	else
	{
		//dbfreebuf(self->dbproc);
		return Py_BuildValue("i",0);
	}
}





// select_db
static PyObject *_mssql_select_db(PY_DBPROC *self, PyObject *args)
{
	char *Dbname;
	if(!PyArg_ParseTuple(args,"s",&Dbname))
	{
		PyErr_SetString(_mssql_error,"Usage: selected_db(Dbname)");
		return NULL;
	}

	if(!self->connected)
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}

#ifdef MS_WIN32
	ZeroMemory(_mssql_query_error,MSSQLDB_MSGSIZE);
	ZeroMemory(_mssql_query_message,MSSQLDB_MSGSIZE);
#else
	memset(_mssql_query_error,0,MSSQLDB_MSGSIZE);
	memset(_mssql_query_message,0,MSSQLDB_MSGSIZE);
#endif

	if(dbuse(self->dbproc,Dbname))
	{
		return Py_BuildValue("i",1);
	}
	else
	{
		return Py_BuildValue("i",0);
	}
}



// close
static PyObject *_mssql_close(PY_DBPROC *self, PyObject *args)
{
	if(!PyArg_NoArgs(args))
	{
		PyErr_SetString(_mssql_error, "This Method is not required Arguments");
		return NULL;
	}
	if(self->connected)
	{
#ifdef MS_WIN32
	ZeroMemory(_mssql_query_error,MSSQLDB_MSGSIZE);
	ZeroMemory(_mssql_query_message,MSSQLDB_MSGSIZE);
#else
	memset(_mssql_query_error,0,MSSQLDB_MSGSIZE);
	memset(_mssql_query_message,0,MSSQLDB_MSGSIZE);
#endif

		dbfreebuf(self->dbproc);
#ifdef MS_WIN32
		if(0==dbclose(self->dbproc))
		{
			PyErr_SetString(_mssql_error,"Fail the DB Connection free and close");
			return NULL;
		}
#else
		dbclose(self->dbproc);
#endif
		self->connected=0;
	}
	else
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}



// error
static PyObject *_mssql_errmsg(PY_DBPROC *self, PyObject *args)
{
	if(!PyArg_NoArgs(args))
	{
		PyErr_SetString(_mssql_error, "This Method is not required Arguments");
		return NULL;
	}
	if(!self->connected)
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}
	return Py_BuildValue("s",_mssql_query_error);
}


// message
static PyObject *_mssql_stdmsg(PY_DBPROC *self, PyObject *args)
{
	if(!PyArg_NoArgs(args))
	{
		PyErr_SetString(_mssql_error, "This Method is not required Arguments");
		return NULL;
	}
	if(!self->connected)
	{
		PyErr_SetString(_mssql_error,"Not Connected DB Connection");
		return NULL;
	}
	return Py_BuildValue("s",_mssql_query_message);
}


// PY_DBPROC ޼ҵ
static PyMethodDef PY_DBPROC_methods[] = {
	{"select_db",(PyCFunction)_mssql_select_db,1},
	{"query",(PyCFunction)_mssql_query,1},
	{"fetch_array",(PyCFunction)_mssql_fetch_array,0},
	{"close",(PyCFunction)_mssql_close,0},
	{"errmsg",(PyCFunction)_mssql_errmsg,0},
	{"stdmsg",(PyCFunction)_mssql_stdmsg,0},
	{NULL,NULL}
};

// PY_DBPROC dealloc
static void PY_DBPROC_dealloc(PY_DBPROC *self)
{
	PyObject *o;
	if(self->connected)
	{
		o =  _mssql_close(self, NULL);
		Py_XDECREF(o);
	}
	PyMem_Free((char *) self);
}

// PY_DBPROC getattr
static PyObject * PY_DBPROC_getattr(PY_DBPROC *self, char *name)
{
	PyObject *res;
	res = Py_FindMethod(PY_DBPROC_methods, (PyObject *)self, name);
	if(res != NULL)
		return res;
	PyErr_Clear();
	if (strcmp(name, "connected") == 0)
		return PyInt_FromLong((long)(self->connected));
	PyErr_SetString(_mssql_error,"UnKnown method");
	return NULL;
}

// PY_DBPROC repr
static PyObject * PY_DBPROC_repr(PY_DBPROC *self)
{
	char buf[300];
	if(self->connected)
		sprintf(buf,"<Opened mssql connection at %lx>",(long)self);
	else
		sprintf(buf,"<Closed mssql connection at %lx>",(long)self);
	return PyString_FromString(buf);
}



// PY_DBPROC 
PyTypeObject PY_DBPROC_Type = {
	#ifndef MS_WIN32
		PyObject_HEAD_INIT(&PyType_Type)
	#else
		PyObject_HEAD_INIT(NULL)
	#endif
	0,
	"PDBPROCESS",
	sizeof(PY_DBPROC),
	0,
	(destructor)PY_DBPROC_dealloc,
	0,
	(getattrfunc)PY_DBPROC_getattr,
	0,
	0,
	(reprfunc)PY_DBPROC_repr,
};













int DetermineRowSize
    (
    DBPROCESS* dbproc,  // The SQL Server connection structure
    int col             // column size to get, 0 for all
    )
    {
    int x,cols=0;         // counters
    int length=0;       // total length of column(row).
    int namelength;     // length of name of column
    int prlength;       // printable length
    char *name;         // pointer to column name
    
    if (!col)           // get number of columns
        cols = dbnumcols(dbproc);
    
    // count from 1 to numcols if col is 0, else x will = col only
    for(x=((col) ? col : 1);x<=((col) ? col : cols);x++)
        {
        switch(dbcoltype(dbproc,x)) // get column type, determine SQLCHAR
            {                       // converted length
            case SQLNUMERIC:
            case SQLDECIMAL:
                {
			#ifdef MS_WIN32
                DBCOL Col;

                Col.SizeOfStruct = sizeof(DBCOL);

                dbcolinfo(dbproc, CI_REGULAR, x, 0, &Col);
                prlength = Col.Precision + 2;
			#else
                prlength = dbcollen(dbproc,x)+2;
			#endif
                }
                break;

            case SQLBIT:            // The PR... values are found in the
                //prlength = PRBIT;   // SQLDB.H header file.
                prlength = 3;   // SQLDB.H header file.
                break;
            case SQLINT1:
                //prlength = PRINT1;
                prlength = 3;
                break;
            case SQLINT2:
                //prlength = PRINT2;
                prlength = 6;
                break;
            case SQLINT4:
                //prlength = PRINT4;
                prlength = 11;
                break;
            case SQLFLT8:
                //prlength = PRFLT8;
                prlength = 20;
                break;
            case SQLDATETIME:
                //prlength = PRDATETIME;
                prlength = 27;
                break;
            case SQLMONEY:
                //prlength = PRMONEY;
                prlength = 26;
                break;
            case SQLVARBINARY:      // VARBINARY IMAGE, and BINARY
            case SQLBINARY:         // convert to 2 times length
            case SQLIMAGE:
                prlength = dbcollen(dbproc,x)*2;
                break;
            default :
                prlength = dbcollen(dbproc,x);  // other types are maximum of
                break;                          // actual column length
            }

        name = (char*) dbcolname(dbproc, x);    // names may be longer than
        namelength = (name) ? strlen(name) : 0; // column so use name len if

        if (prlength < namelength)      // longer of two.
            length += namelength + 1;   // add one for space between
        else                            // columns
            length += prlength + 1;
        }

    return length;              // return the length of the field
    }

PyObject *GetHeaders ( DBPROCESS *dbproc )
{
    int x,cols;//,size;        // counters
	PyObject *headerSet=NULL;

    //size = DetermineRowSize(dbproc,0);  // get size of buffers
    //ptr = colnames = malloc(size+1);    // get name buffer
    //memset (colnames,' ',size);
    
    cols = dbnumcols(dbproc);           // get number of columns
	if (!(headerSet = PyTuple_New(cols)))
	{
		PyErr_SetString(_mssql_error,"Not Created Column Tuple");
		return NULL;
	}

    for(x = 1; x <= cols; x++)          // loop on all columns
    {
        //size = DetermineRowSize(dbproc,x);      // get size of this column
        //colname = (char *)dbcolname(dbproc,x);  // get column name
        //strncpy(ptr,colname,strlen(colname));   // copy name
		PyTuple_SET_ITEM(headerSet,x-1,Py_BuildValue("s",(char *)dbcolname(dbproc,x)));
        //ptr+=size;                              // move to next position
    }

    //*ptr = '\0';                // null term both strings
    //printf("%s\n",colnames);    // print both strings
    //free(colnames);             // free both buffers

    return headerSet;
}

PyObject *GetRow ( DBPROCESS *dbproc)
{
    int x,cols,coltype;  // counters
    char *data;             // column data pointer
	int size,datasize;
	char *datavals;
	long *datavali,intdata;
	double *datavalf,ddata;
	PyObject *record;

	datavali=&intdata;
	datavalf=&ddata;
    
    cols = dbnumcols(dbproc);               // get number of columns
	if (!(record = PyTuple_New(cols)))
	{
		PyErr_SetString(_mssql_error,"Not Created Record Tuple");
		return NULL;
	}
    
    for(x=1;x<=cols;x++)                    // do all columns
    {
        coltype = dbcoltype(dbproc,x);
		size = DetermineRowSize(dbproc,x);
        data = (BYTE *)dbdata(dbproc,x);    // get pointer to column's data
        if(data == NULL)            // if NULL, use "NULL"
        {
			PyTuple_SET_ITEM(record,x-1,Py_BuildValue(""));
        }
        else                                // else have data, so convert to char
        {
			datasize=dbdatlen(dbproc,x);
			if (datasize < size && (coltype == SQLNUMERIC || 
									coltype == SQLDECIMAL || 
									coltype == SQLINT1 || 
									coltype == SQLINT2 ||
									coltype == SQLINT4 ||
									coltype == SQLFLT8 || 
									coltype == SQLFLT4))
			{
				if(coltype==SQLFLT4 || coltype==SQLFLT8)
				{
					dbconvert(dbproc,coltype,data,datasize, SQLFLT8,(LPBYTE)datavalf,(DBINT)size-1);
					PyTuple_SET_ITEM(record,x-1,Py_BuildValue("d",*datavalf));
				}
				else
				{
					dbconvert(dbproc,coltype,data,datasize, SQLINT4,(LPBYTE)datavali,(DBINT)size-1);
					PyTuple_SET_ITEM(record,x-1,Py_BuildValue("i",*datavali));
				}
			}
			else
			{
				if(coltype==SQLDATETIME || coltype==SQLDATETIME4) {
					DBDATEREC di;
					char strdi[25]={0,};
					dbdatecrack(dbproc, &di, (DBDATETIME*)data);
#ifdef MS_WIN32
					sprintf(strdi, "%02d-%02d-%02d %02d:%02d:%02d", 
							di.year,
							di.month+1,
							di.day,
							di.hour,
							di.minute,
							di.second);
							//di.millisecond);
#else
					sprintf(strdi, "%02d-%02d-%02d %02d:%02d:%02d", 
							di.dateyear,
							di.datemonth+1,
							di.datedmonth,
							di.datehour,
							di.dateminute,
							di.datesecond);
							//di.datemsecond);
#endif
					PyTuple_SET_ITEM(record,x-1,Py_BuildValue("s",strdi));
				} else {
					datavals=malloc(datasize+1);
#ifdef MS_WIN32
					ZeroMemory(datavals,datasize+1);
#else
					memset(datavals,0,datasize+1);
#endif
					memmove(datavals,data,datasize);
			//		printf("str %d\t- %s\n",datasize,datavals);
					PyTuple_SET_ITEM(record,x-1,Py_BuildValue("s",datavals));
					free(datavals);
				}
			}
        }
    }

    return record;             // done
}




